#pragma once
#define enable_console true
#include <csignal>
#if enable_console
#include <cstdio>
#endif
#include "array_view.h"
#include <array>
#include <cstdlib>
#include <string>
#include <vector>
#include <tuple>
using std::string;
using std::to_string;
typedef unsigned uint; //clang and VSCode don't do this automatically

using std::string_view;

void o(string_view s) {
#if enable_console
	fwrite(s.data(), sizeof(char), s.size(), stdout);
	//C++ iostreams suck with emscripten
#elif _MSC_VER
#pragma comment(linker, "/SUBSYSTEM:windows /ENTRY:mainCRTStartup")
#endif
	//EM_ASM_({console.log(UTF8ToString($0))}, s.c_str());
}
#if _MSC_VER
inline void __builtin_unreachable() {}
#endif

template <typename T>
auto o(const T& s) -> decltype(to_string(std::declval<T>()), void()) {
	string str = to_string(s);
	o(string_view(str));
}

constexpr char no_whitespace_before[] = {'@', '/', '\\', '?', '!', '.', ',', ';', ':', ')', ']', '}', '\'', '\"', '\r', '\n', '\t'};
constexpr char no_whitespace_after[] = {'\r', '\n', '\t', '{', '[', '(', '#', '$', '/', '\\', ' '}; //having ' ' in there means zero spaces to auto-space, one space to force a space, two spaces to create an extra space. forcing spaces is useful for arbitrary symbols
template <typename T>
bool whitespace_before(const T& whatever) {
	if constexpr (std::is_convertible_v<T, std::string_view>) {
		string_view str = whatever;
		if (str.empty()) return false;
		for (char n : no_whitespace_before)
			if (str.front() == n) return false;
	}
	//anything not already in text form returns true
	return true;
}

template <typename T>
bool whitespace_after(const T& whatever) {
	if constexpr (std::is_convertible_v<T, string_view>) {
		string_view str = whatever;
		if (str.empty()) return false;
		for (char n : no_whitespace_after)
			if (str.back() == n) return false;
	}
	return true;
}

//rank is used to order template choosing. https://stackoverflow.com/a/34419541
//insert space is true if the previous word didn't have a space after it.
template <unsigned int N>
struct rank : rank<N - 1> {};
template <>
struct rank<0> {};

inline void outc_internal(rank<0>, bool insert_space) {}

template <typename Arg, typename... Args>
inline void outc_internal(rank<0>, bool insert_space, Arg arg, Args&&... args) {
	if (insert_space && whitespace_before(arg))
		o(" ");
	o(arg);
	outc_internal(rank<1>{}, whitespace_after(arg), std::forward<Args>(args)...);
}

template <typename T1, typename T2, typename... Args>
inline void outc_internal(rank<1>, bool insert_space, std::pair<T1, T2>& arg, Args&&... args) {
	outc_internal(rank<1>{}, insert_space, arg.first);
	outc_internal(rank<1>{}, true, arg.second);
	outc_internal(rank<1>{}, true, std::forward<Args>(args)...);
}

//https://stackoverflow.com/a/41171552
template <class TupType, size_t... I>
inline void detupleize(bool insert_space, const TupType& _tup, std::index_sequence<I...>) {
	(..., (outc_internal(rank<1>{}, (insert_space || I), std::get<I>(_tup))));
}

template <class... T, typename... Args>
inline void outc_internal(rank<1>, bool insert_space, const std::tuple<T...>& tup, Args&&... args) {
	detupleize(insert_space, tup, std::make_index_sequence<sizeof...(T)>());
	outc_internal(rank<1>{}, true, std::forward<Args>(args)...);
}

#ifndef _MSC_VER
template <typename T, typename... Args, class = typename std::enable_if_t<std::is_convertible_v<T, array_view1d<typename T::value_type>>>>
inline void outc_internal(rank<1>, bool insert_space, T pre_arg, Args&&... args) {
	auto arg = array_view1d<typename T::value_type>(pre_arg);
	bool first = true;
	for (auto& x : arg) {
		if (first) {
			outc_internal(rank<1>{}, insert_space, x);
			first = false;
		}
		else {
			o(" ");
			outc_internal(rank<1>{}, false, x); //force whitespace
		}
	}
	outc_internal(rank<1>{}, true, std::forward<Args>(args)...);
}
#endif
template <typename... Args>
void outc(Args&&... args) {
	outc_internal(rank<1>{}, false, std::forward<Args>(args)...);
	o("\n"); //strings only print in emscripten when they end in \n
}

//noreturn silences llvm warning about control reaching end without returning
template <typename... Args>
[[noreturn]] inline void error(Args&&... args) {
	outc(std::forward<Args>(args)...);
#ifndef _MSC_VER
	__builtin_trap(); //used to lose frame information when compiled with clang, but seems fixed with clang 6.0. zero frames added
	__builtin_unreachable(); //silence noreturn warnings
#else
	//abort(); //adds two frames to the debugger
	std::raise(SIGABRT); //1 frame added
	//throw; //adds a ton of stack frames
#endif
	//assert(false); //adds a huge number of useless stack frames
}

template <typename... Args>
inline void check(bool b, Args&&... args) {
	if (!b) error(std::forward<Args>(args)...);
}

#ifdef NDEBUG
template <typename... Args>
[[noreturn]] inline void error_assert(Args&&... args) {
#ifndef _MSC_VER
	__builtin_unreachable();
#endif
}
#else
template <typename... Args>
[[noreturn]] inline void error_assert(Args&&... args) {
	error(std::forward<Args>(args)...);
}
#endif
template <typename... Args>
inline void check_assert(bool b, Args&&... args) {
	if (!b) error_assert(std::forward<Args>(args)...);
}