//include this file only once, ever. it's a cpp file. that's because it calls timeBeginPeriod on Windows
//this has been tested to give extremely good accuracy on Windows, up to about 1us.
//this has not been tested at all on Linux.
#define BENCHMARK_SLEEP false

#include <chrono>
#include <thread>

#if _WIN32
#include <windows.h>
//blank line to prevent clang-format from re-ordering headers
#include <Mmsystem.h>
#pragma comment(lib, "winmm.lib")
bool improve_timer_resolution_on_Windows = []() {
	timeBeginPeriod(1);
	return 0;
}();
uint64_t expected_resolution = 1000000; //1ms on Windows
#else
uint64_t expected_resolution = 1; //not sure what it should be on Linux
#endif

#if BENCHMARK_SLEEP
bool benchmark_low_sleeps = []() {
	outc("testing sleep accuracy, all counts in microseconds:");
	using frame_clock = std::chrono::steady_clock;
	using microsecond = std::chrono::microseconds;

	for (int x = 0; x < 20; ++x) {
		auto sleep_time = microsecond(200 * x);
		auto now = frame_clock::now().time_since_epoch();
		std::this_thread::sleep_for(sleep_time);
		auto end = frame_clock::now().time_since_epoch();

		outc("asked for", sleep_time.count(), "actual wait", std::chrono::duration_cast<microsecond>(end - now).count());
		//outc("frame delay ", float(frame_clock::period::den) / (time_frame_end - time_frame_start));

		/*
		gcc linux uses nanosleep for std::this_thread::sleep_for(): http://www.martani.net/2011/07/nanosleep-usleep-and-sleep-precision.html
		clock_nanosleep alternative: https://stackoverflow.com/questions/7794955/why-is-clock-nanosleep-preferred-over-nanosleep-to-create-sleep-times-in-c
		https://stackoverflow.com/questions/3523442/difference-between-clock-realtime-and-clock-monotonic
		http://man7.org/linux/man-pages/man2/nanosleep.2.html says nanosleep uses MONOTONIC on linux anyway so I don't have to worry about this
		the spec says POSIX.1 specification for clock_settime(2) says that discontinuous changes in CLOCK_REALTIME should not affect nanosleep()
		so Linux is semi-defying the spec, although the spec also says that end_time changes should not affect sleep times. so really, linux is fine and the spec is stupid.
		*/
	}
	return 0;
}();
#endif

//spinwait sleep for slightly more than than the asked-for period, but basically equal. uses a monotonic clock, so good for small and accurate timepoints, but not for long periods
void accurate_sleep_until(uint64_t end_time, uint64_t current_time = 0) {
	using frame_clock = std::chrono::steady_clock;
	auto get_current_time = []() { return frame_clock::now().time_since_epoch().count(); };
	static uint64_t sum_zero_sleep_times = 0;
	static int64_t zero_sleep_tries = 0;
	static int64_t average_sleep_duration = 0;
	if (sum_zero_sleep_times > 100000000) {
		average_sleep_duration = sum_zero_sleep_times / zero_sleep_tries;
		//outc("fastsleep average", sum_zero_sleep_times / zero_sleep_tries); //note: writing to console will create oversleeps
		zero_sleep_tries = 0;
		sum_zero_sleep_times = 0;
	}

	if (current_time == 0) current_time = get_current_time();
	//if (end_time <= current_time - average_sleep_duration) return;
	if (end_time <= current_time) return; //I feel that subtracting average_sleep_duration is not worth it

	if (end_time - current_time > expected_resolution) {
		std::this_thread::sleep_for(frame_clock::duration(end_time - current_time - expected_resolution)); //- expected_resolution * 2 / 3 gives some misses on Windows
		//outc("asked", end_time - current_time - expected_resolution , get_current_time() - startaa);
	}
/*
	uint64_t end_time;
	while (1) {
#if BENCHMARK_SLEEP
		uint64_t start_time = get_current_time();
#endif
#if __linux__
		//don't sleep. //https://stackoverflow.com/a/11049590
#else
		std::this_thread::yield(); //700ns average. when there's activity (which is often), this sucks and goes way over.
#endif
		//std::this_thread::sleep_for(std::chrono::seconds(0)); //50ns average
		//std::this_thread::sleep_for(std::chrono::microseconds(1)); //on Windows, there's a difference between sleeping for 0 and for 1, and no difference for 1 - 999us. but it's only 10ns. so probably sleep(1) is doing some extra calculation.
		end_time = get_current_time(); //note: I could estimate how many sleeps are left by keeping track of the average sleep end_time. this would reduce the number of timer calls. not sure if that is useful for decreasing CPU usage.

#if BENCHMARK_SLEEP
		++zero_sleep_tries;
		sum_zero_sleep_times += end_time - start_time;
#endif
		if (end_time > end_time - expected_resolution / 2) break;
	}*/

	while (get_current_time() < end_time) {}

	//uint64_t result_time = get_current_time();
	//if (result_time > end_time + 10000) outc("waited too long in nanoseconds:", result_time - end_time);
}